use core::future::Future;

use embassy_rp::usb::Driver;
use embassy_rp::{interrupt, peripherals::USB};

use embassy_usb::class::hid::{HidWriter, ReportId, RequestHandler, State};
use embassy_usb::control::OutResponse;
use embassy_usb::{Builder, Config, UsbDevice};

use usbd_hid::descriptor::{MouseReport, SerializedDescriptor};

pub struct MouseConfig<'a> {
    pub vendor_id: u16,
    pub product_id: u16,
    pub manufacturer: &'a str,
    pub product: &'a str,
    pub serial_number: &'a str,
}

pub struct MouseHID<'d> {
    writer: HidWriter<'d, Driver<'d, USB>, 5>,
    usb: UsbDevice<'d, Driver<'d, USB>>,

    state: State<'d>,

    device_descriptor: [u8; 256],
    config_descriptor: [u8; 256],
    bos_descriptor: [u8; 256],
    control_buf: [u8; 64],
}

impl<'d> MouseHID<'d> {
    pub async fn new(usb: USB, mouse_config: MouseConfig<'_>) -> MouseHID<'d> {
        // Create the driver, from the HAL.
        let irq = interrupt::take!(USBCTRL_IRQ);
        let driver = Driver::new(usb, irq);

        let mut config = Config::new(mouse_config.vendor_id, mouse_config.product_id);
        config.manufacturer = Some(mouse_config.manufacturer);
        config.product = Some(mouse_config.product);
        config.serial_number = Some(mouse_config.serial_number);
        config.device_class = 0;
        config.device_protocol = 2;

        let mut device_descriptor = [0; 256];
        let mut config_descriptor = [0; 256];
        let mut bos_descriptor = [0; 256];
        let mut control_buf = [0; 64];

        let request_handler = MyRequestHandler {};

        let mut state = State::new();

        let mut builder = Builder::new(
            driver,
            config,
            &mut device_descriptor,
            &mut config_descriptor,
            &mut bos_descriptor,
            &mut control_buf,
        );

        // Create classes on the builder.
        let config = embassy_usb::class::hid::Config {
            report_descriptor: MouseReport::desc(),
            request_handler: Some(&request_handler),
            poll_ms: 60,
            max_packet_size: 8,
        };

        let writer = HidWriter::<_, 5>::new(&mut builder, &mut state, config);

        // Build the builder.
        let usb = builder.build();

        MouseHID {
            writer,
            usb,

            state,

            device_descriptor,
            config_descriptor,
            bos_descriptor,
            control_buf,
        }
    }

    pub async fn run(&'d mut self) -> impl Future<Output = !> + 'd {
        self.usb.run()
    }

    pub async fn mouse_report(&mut self, report: MouseReport) {
        match self.writer.write_serialize(&report).await {
            Ok(()) => {}
            Err(_e) => (),
        }
    }
}

struct MyRequestHandler {}

impl RequestHandler for MyRequestHandler {
    fn get_report(&self, id: ReportId, _buf: &mut [u8]) -> Option<usize> {
        None
    }

    fn set_report(&self, id: ReportId, data: &[u8]) -> OutResponse {
        OutResponse::Accepted
    }

    fn set_idle_ms(&self, id: Option<ReportId>, dur: u32) {}

    fn get_idle_ms(&self, id: Option<ReportId>) -> Option<u32> {
        None
    }
}
